import k8s from "@kubernetes/client-node";
import {
  getDeployment,
  getDeployments,
  getServerAssets,
  scaleDeployment,
} from "./k8s-server-control.js";
import { ERR } from "../util/logging.js";
const kc = new k8s.KubeConfig();
kc.loadFromDefault();

const k8sMetrics = new k8s.Metrics(kc);
const namespace = process.env.MCL_SERVER_NAMESPACE;

// Gets the all assets for the server
export async function getServer(req, res) {
  const serverSpec = req.body;
  if (!serverSpec) return res.sendStatus(400);
  if (!serverSpec.name) return res.status(400).send("Server name required!");
  const { name } = serverSpec;
  getServerAssets(name)
    .then((server) => res.status(200).json(server))
    .catch((e) => res.status(500).send(e));
}

export async function startServer(req, res) {
  const serverSpec = req.body;
  if (!serverSpec) return res.sendStatus(400);
  if (!serverSpec.name) return res.status(400).send("Server name required!");
  const { name } = serverSpec;
  try {
    await scaleDeployment(name, true);
    res.sendStatus(200);
  } catch (e) {
    ERR("SERVER CONTROL", e);
    res.status(500).send(`Error updating server '${name}'!`);
  }
}

export async function stopServer(req, res) {
  const serverSpec = req.body;
  if (!serverSpec) return res.sendStatus(400);
  if (!serverSpec.name) return res.status(400).send("Server name required!");
  const { name } = serverSpec;
  try {
    await scaleDeployment(name, false);
    res.sendStatus(200);
  } catch (e) {
    ERR("SERVER CONTROL", e);
    res.status(500).send(`Error updating server '${name}'!`);
  }
}

export function serverList(req, res) {
  getDeployments()
    .then((sd) => res.json(sd.map((s) => s.metadata.name.substring(4))))
    .catch((e) => {
      ERR("SERVER CONTROL", e);
      res.status(500).send("Couldn't get server list");
    });
}

export async function getServers(req, res) {
  const serverDeployments = await getDeployments();
  const podMetricsResponse = await k8sMetrics.getPodMetrics(namespace);

  var name, metrics, started;
  const servers = serverDeployments.map((s) => {
    name = s.metadata.name.substring(4);
    metrics = null;
    started = !!s.spec.replicas;
    const pod = podMetricsResponse.items.find(({ metadata: md }) => {
      return md.labels && md.labels.app && md.labels.app === `mcl-${name}-app`;
    });
    if (pod) {
      const podCpus = pod.containers.map(
        ({ usage }) => parseInt(usage.cpu) / 1_000_000,
      );
      const podMems = pod.containers.map(
        ({ usage }) => parseInt(usage.memory) / 1024,
      );
      metrics = {
        cpu: Math.ceil(podCpus.reduce((a, b) => a + b)),
        memory: Math.ceil(podMems.reduce((a, b) => a + b)),
      };
    }

    return { name, metrics, started };
  });
  var clusterMetrics = { cpu: 0, memory: 0 };
  if (servers.length > 1) {
    const clusterCpu = servers
      .map(({ metrics }) => (metrics ? metrics.cpu : 0))
      .reduce((a, b) => a + b);
    const clusterMem = servers
      .map(({ metrics }) => (metrics ? metrics.memory : 0))
      .reduce((a, b) => a + b);
    clusterMetrics = { cpu: clusterCpu, memory: clusterMem };
  }
  res.json({ servers, clusterMetrics });
}
